--[[
Actor class 
===========

DESCRIPTION
-----------

class for all moving objects

GetActors などでとれる `id' のラッパクラスです。
GetV や、その他もろもろのユーティリティ関数も含んでいます。

プレイヤーやモンスターどちらに対しても使えます。


SYNOPSYS
--------


どんなメソッドがあるかはソース読んだほうが早いと思われます。

また  [mobdata.ini]( http://himenomikoto.at.infoseek.co.jp/kulus_project/#DB )
を使用し、詳細なモンスターデータも取得できます。

	monster = Actor.new(monster_id)
	target:getMonsterInfo() -- これの呼びだしが必要
	monster.info.NAME

### クラスメソッド

 * Actor.all
   GetActors を実行し、
   Actor インスタンスでラップした
   List インスタンスを返す

 * Actor.monsters
   all からモンスターのみとりだした List

 * Actor.players
   all からプレイヤーのみとりだした List

距離 10 以下のモンスターを取得

	local pick = Actor.monsters():select(function (a)
		if self.owner:distanceFrom(a) < 10 then
			return true
		end
	end)

画面内のプレイヤーの職業を列挙

	Actor.players():each(function (a)
		d(a:jobName())
	end)

NOTE
----

ホムンクルスのスキル使用などは Homun クラスに定義されています。
ホムンクルスのインスタンスを生成するときには Homun を使ってください。

プレイヤーとモンスターをサブクラスにしていないのは、使いやすさのためです。

]]

require "Class"
require "List"
require "Try"
require "INI"
require "RO"

Actor = Class {
	JOBNAME1 = {
		'Novice',
		'Swordman',
		'Magician',
		'Archer',
		'Acolyte',
		'Merchant',
		'Thief',
		'Knight',
		'Priest',
		'Wizard',
		'Blacksmith',
		'Hunter',
		'Assassin',
		'Knight (rider)',
		'Crusader',
		'Monk',
		'Sage',
		'Rogue',
		'Alchemist',
		'Bard',
		'Dancer',
		'Crusader (rider)',
		'Wedding',
		'Super Novice',
		'Gunslinger',
		'Ninja'
	},

	JOBNAME2 = {
		'Novice High',
		'Swordman High',
		'Magician High',
		'Archer High',
		'Acolyte High',
		'Merchant High',
		'Thief High',
		'Lord Knight',
		'High Priest',
		'High Wizard',
		'White Smith',
		'Sniper',
		'Assassin Cross',
		'Lord Knight (rider)',
		'Paladin',
		'Champion',
		'Professor',
		'Chaser',
		'Creator',
		'Clown',
		'Gypsy',
		'Paladin (rider)',
		'Baby',
		'Baby Swordman',
		'Baby Magician',
		'Baby Archer',
		'Baby Acolyte',
		'Baby Merchant',
		'Baby Thief',
		'Baby Knight',
		'Baby Priest',
		'Baby Wizard',
		'Baby Blacksmith',
		'Baby Hunter',
		'Baby Assassin',
		'Baby Knight (rider)',
		'Baby Crusader',
		'Baby Monk',
		'Baby Sage',
		'Baby Rogue',
		'Baby Alchemist',
		'Baby Bard',
		'Baby Dancer',
		'Baby Crusader (rider)',
		'Super Baby',
		'TaeKwon',
		'Star Gladiator',
		'Star Gladiator',
		'Soul Linker'
	},

	__eq = function (op1, op2)
		return op1.id == op2.id
	end,

	initialize = function (self, id)
		self.id = id
	end,

	x = function (self)
		local x, y = self:position();
		return x
	end,

	y = function (self)
		local x, y = self:position();
		return y
	end,

	position = function (self)
		local x, y =  GetV(V_POSITION, self.id)
		if (x == -1 or y == -1) then
			error("V_POSITION return -1; Invalid actor id?")
		end
		return x, y
	end,

	exists = function (self)
		if self.id == 0 then return false end
		return Actor.all():isInclude(self)
	end,

	motion = function (self)
		return GetV(V_MOTION, self.id)
	end,

	attackRange = function (self)
		return GetV(V_ATTACKRANGE, self.id)
	end,

	target = function (self)
		return Actor.new(GetV(V_TARGET, self.id))
	end,

	skillAttackRange = function (self)
		return GetV(V_SKILLATTACKRANGE, self.id)
	end,

	hp = function (self)
		return GetV(V_HP, self.id)
	end,

	sp = function (self)
		return GetV(V_SP, self.id)
	end,

	maxHp = function (self)
		local ret = GetV(V_MAXHP, self.id)
		if ret == -1 then
			self:getMonsterInfo()
			return tonumber(self.info.MAXHP)
		else
			return ret
		end
	end,

	maxSp = function (self)
		return GetV(V_MAXSP, self.id)
	end,

	-- それぞれ % で返す
	hpRate = function (self)
		return self:hp() / self:maxHp() * 100
	end,

	spRate = function (self)
		return self:sp() / self:maxSp() * 100
	end,

	merType = function (self)
		return GetV(V_MERTYPE, self.id)
	end,

	homunType = function (self)
		return GetV(V_HOMUNTYPE, self.id)
	end,

	isMonster = function (self)
		return (IsMonster(self.id) == 1)
	end,

	-- self から見た other の方向
	directionTo = function (self, other)
		local sx, sy = self:position()
		local tx, ty = other:position()
		local rad = math.atan2(-(ty - sy), tx - sx)
		local deg = rad * 180 / math.pi + 90
		if deg < 0 then deg = deg + 360 end
		return math.mod(deg, 360)
	end,

	-- http://himenomikoto.at.infoseek.co.jp/kulus_project/playerjobdata.lua
	isPlayer = function (self)
		if self.id > 100000 then
			local ty = self:homunType()
			if 0 <= ty and ty <= 25 then
				return true
			end
			if 4001 <= ty and ty <= 4049 then
				return true
			end
			return false
		else
			return false
		end
	end,

	isPlant = function (self)
		local ty = self:homunType()
		for i, v in ipairs({1078, 1079, 1080, 1081, 1082, 1083}) do
			if ty == v then return true end
		end
		return false
	end,

	isBioPlant = function (self)
		local ty = self:homunType()
		for i, v in ipairs({1555, 1575, 1579, 1589, 1590}) do
			if ty == v then return true end
		end
		return false
	end,

	isMarineSphere = function (self)
		return (self:homunType() == 1142)
	end,

	isEgg = function (self)
		local ty = self:homunType()
		for i, v in ipairs({1008, 1047, 1097, 1236, 1142, 1721}) do
			if ty == v then return true end
		end
		return false
	end,

	jobName = function (self)
		if self:isPlayer() then
			local ty = self:homunType()
			if 0 <= ty and ty <= 25 then
				return self.JOBNAME1[ty+1]
			end
			if 4001 <= ty and ty <= 4049 then
				return self.JOBNAME2[ty-4000]
			end
		else
			return "Not Player"
		end
	end,

	-- monster info
	--
	getMonsterInfo = function (self)
		if not self.info then
			Actor.readMobDB()
			self.info = Actor.db[tostring(self:homunType())] or {}
		end
	end,

	race = function (self)
		self:getMonsterInfo()
		if self.info.race == 0 then return "none" end
		if self.info.race == 1 then return "undead" end
		if self.info.race == 2 then return "brute" end
		if self.info.race == 3 then return "plant" end
		if self.info.race == 4 then return "insect" end
		if self.info.race == 5 then return "fish" end
		if self.info.race == 6 then return "demon" end
		if self.info.race == 7 then return "human" end
		if self.info.race == 8 then return "angel" end
		if self.info.race == 9 then return "dragon" end
	end,

	attr = function (self)
		self:getMonsterInfo()
		if self.info.attr == 0 then return "none" end
		if self.info.attr == 1 then return "water" end
		if self.info.attr == 2 then return "earth" end
		if self.info.attr == 3 then return "fire" end
		if self.info.attr == 4 then return "wind" end
		if self.info.attr == 5 then return "poison" end
		if self.info.attr == 6 then return "holy" end
		if self.info.attr == 7 then return "shadow" end
		if self.info.attr == 8 then return "ghost" end
		if self.info.attr == 9 then return "undead" end
	end,

	size = function (self)
		self:getMonsterInfo()
		if self.info.attr == 0 then return "small" end
		if self.info.attr == 1 then return "middle" end
		if self.info.attr == 2 then return "large" end
	end,

	isSmall = function (self)
		self:getMonsterInfo()
		return self.info.size == 0
	end,

	isMiddle = function (self)
		self:getMonsterInfo()
		return self.info.size == 1
	end,

	isLarge = function (self)
		self:getMonsterInfo()
		return self.info.size == 2
	end,

	-- places
	distanceFrom = function (self, other, y)
		local x1, y1 = GetV(V_POSITION, self.id)
		local x2, y2 = GetV(V_POSITION, other.id)
		if y then
			x2, y2 = other, y
		else
			x2, y2 = GetV(V_POSITION, other.id)
		end

		if (x1 == -1 or x2 == -1) then return 9999 end
		return math.floor(math.sqrt((x1 - x2) ^ 2 + (y1 - y2) ^ 2))
	end,

	isInSight = function (self, other)
		local d = self:distanceFrom(other)
		return (d < 10)
	end,

	isInAttackSight = function (self, other)
		local d = self:distanceFrom(other)
		local a = self:attackRange()
		return (a >= d)
	end,

	-- good methods for motion
	isStand = function (self)
		return (self:motion() == MOTION_STAND)
	end,

	isMove = function (self)
		return (self:motion() == MOTION_MOVE)
	end,

	isAttack = function (self)
		local motion = self:motion()
		return (motion == MOTION_ATTACK or motion == MOTION_ATTACK2)
	end,

	isDead = function (self)
		return (self:motion() == MOTION_DEAD)
	end,

	isDamage = function (self)
		return (self:motion() == MOTION_DAMAGE)
	end,

	isBenDown = function (self)
		return (self:motion() == MOTION_BENDOWN)
	end,

	isSit = function (self)
		return (self:motion() == MOTION_SIT)
	end,

	isSkill = function (self)
		local motion = self:motion()
		return (motion == MOTION_SKILL or
		        motion == MOTION_SPIRAL or
		        motion == MOTION_TOSS or
		        motion == MOTION_COUNTER or
		        motion == MOTION_BIGTOSS or
		        motion == MOTION_DESPERADO)
	end,

	isCasting = function (self)
		return (self:motion() == MOTION_CASTING)
	end,

	isCounter = function (self)
		return (self:motion() == MOTION_COUNTER)
	end,

	isPerfom = function (self)
		return (self:motion() == MOTION_PERFORM)
	end,

	isUpper = function (self)
		return (self:motion() == MOTION_UPPER)
	end,

	isDowner = function (self)
		return (self:motion() == MOTION_DOWNER)
	end,

	isSoul = function (self)
		return (self:motion() == MOTION_SOUL)
	end,

	isIn = function (self)
		return (self:motion() == MOTION_IN)
	end,
}

-- return all actors by List
function Actor.all ()
	local actors = GetActors ()
	local ret = List.new()
	for i,v in ipairs(actors) do
		ret:push(Actor.new(v))
	end
	return ret
end

function Actor.players (isIncludeSelf)
	return Actor.all():select(function (i)
		return i:isPlayer()
	end)
end

function Actor.monsters ()
	return Actor.all():select(function (i)
		return i:isMonster()
	end)
end

function Actor.readMobDB ()
	if not Actor.db then
		Actor.db = INI.new():read("AI/USER_AI/mobdata.ini")
		d("Readed mobdata.ini")
	end
end

function Actor.playersByPosition (x, y)
	return Actor.players():select(function (i)
		local ax, ay = i:position()
		if x == ax and y == ay then return true end
	end)
end

function Actor.objectsByPosition (x, y)
	return Actor.all():select(function (i)
		local ax, ay = i:position()
		if x == ax and y == ay then return true end
	end)
end
